# -*- coding: utf-8 -*-

from tuikit.events import Event, Emitter
from tuikit.widget import Widget
from tuikit.common import Coords


class AbstractTableModel(Emitter):
    def __init__(self):
        self.add_events('change', Event)

    ## accessing data

    def get_row(self, row_idx):
        """Get row by index.

        Row is list of items of any type.

        """
        raise IndexError()

    def get_column(self, column_idx):
        """Get column by index (list of items)."""
        raise IndexError()

    def get_item(self, row_idx, column_idx):
        """Get item by row and column index."""
        return self.get_row(row_idx)[column_idx]

    def get_row_header(self, row_idx, header_idx):
        raise IndexError()

    def get_column_header(self, column_idx, header_idx):
        raise IndexError()

    ## filling data

    def set_item(self, row_idx, column_idx, value):
        """Set item by row and column index."""
        self.get_row(row_idx)[column_idx] = value
        self.emit('change')

    def set_row_header(self, row_idx, header_idx, value):
        raise NotImplementedError()

    def set_column_header(self, column_idx, header_idx, value):
        raise NotImplementedError()

    ## reading sizes

    def row_count(self):
        """Get number of rows."""
        return 0

    def column_count(self):
        """Get number of columns."""
        return 0

    def row_header_count(self):
        """Get number of row headers (header columns on left side of table)"""
        return 0

    def column_header_count(self):
        """Get number of column headers (header rows on top of table)"""
        return 0

    ## changing size

    def insert_row(self, row_idx):
        raise NotImplementedError()

    def remove_row(self, row_idx):
        raise NotImplementedError()

    def insert_column(self, column_idx):
        raise NotImplementedError()

    def remove_column(self, column_idx):
        raise NotImplementedError()

    def set_num_headers(self, row_header_count, column_header_count):
        """Set number of rows headers and column headers."""
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()


class TableModel(AbstractTableModel):
    def __init__(self):
        AbstractTableModel.__init__(self)
        # list of lists of items
        self._num_cols = 0
        self._num_col_headers = 0
        self._num_row_headers = 0
        self._rows = []
        self._col_header_rows = []
        self._row_header_rows = []

    ## accessing data

    def get_row(self, row_idx):
        """Get row by index."""
        return self._rows[row_idx]

    def get_column(self, column_idx):
        """Get column by index (list of items)."""
        if column_idx >= self._num_cols or column_idx < 0:
            raise IndexError()
        return [row[column_idx] for row in self._rows]

    def get_row_header(self, row_idx, header_idx):
        return self._row_header_rows[row_idx][header_idx]

    def get_column_header(self, column_idx, header_idx):
        return self._col_header_rows[header_idx][column_idx]

    ## filling data

    def set_row_header(self, row_idx, header_idx, value):
        self._row_header_rows[row_idx][header_idx] = value
        self.emit('change')

    def set_column_header(self, column_idx, header_idx, value):
        self._col_header_rows[header_idx][column_idx] = value
        self.emit('change')

    ## reading sizes

    def row_count(self):
        return len(self._rows)

    def column_count(self):
        return self._num_cols

    def row_header_count(self):
        return self._num_row_headers

    def column_header_count(self):
        return self._num_col_headers

    ## changing size

    def insert_row(self, row_idx):
        row = [None] * self._num_cols
        self._rows.insert(row_idx, row)
        self._row_header_rows.insert(row_idx, [None] * self._num_row_headers)
        self.emit('change')

    def remove_row(self, row_idx):
        del self._rows[row_idx]
        del self._row_header_rows[row_idx]
        self.emit('change')

    def insert_column(self, column_idx):
        for row in self._col_header_rows:
            row.insert(column_idx, None)
        for row in self._rows:
            row.insert(column_idx, None)
        self._num_cols += 1
        self.emit('change')

    def remove_column(self, column_idx):
        for row in self._col_header_rows:
            del row[column_idx]
        for row in self._rows:
            del row[column_idx]
        self._num_cols -= 1
        self.emit('change')

    def set_num_headers(self, row_header_count, column_header_count):
        """Set number of rows headers and column headers."""
        self._num_col_headers = column_header_count
        self._num_row_headers = row_header_count
        self._col_header_rows = []
        for _i in range(self._num_col_headers):
            self._col_header_rows.append([None] * self._num_cols)
        for row in range(len(self._row_header_rows)):
            self._row_header_rows[row] = [None] * self._num_row_headers
        self.emit('change')

    def clear(self):
        self._rows = []
        self._row_header_rows = []
        self._num_cols = 0
        self.set_num_headers(0, 0)


class TableView(Widget):
    def __init__(self, model=None):
        Widget.__init__(self)
        self._default_size.update(20, 20)
        self.allow_focus = True

        # widths of colums, including header columns
        self.column_sizes = []

        self.spacing = 1

        #: Active cell (cursor)
        self.acell = Coords()

        # model
        self._model = None
        self.model = model

        self.add_events('scroll', Event)

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        if self._model:
            self._model.remove_handler('change', self.redraw)
        self._model = value
        if self._model:
            self._model.add_handler('change', self.redraw)
        self._update_sizes()

    def _update_sizes(self):
        for _i in range(self._model.row_header_count()):
            self.column_sizes.append(6)
        for _i in range(self._model.column_count()):
            self.column_sizes.append(8)
        width = sum(self.column_sizes) + self.spacing * (len(self.column_sizes) - 1)
        height = self._model.row_count() + self._model.column_header_count()
        self._default_size.update(w=width, h=height)

    def _draw_column_headers(self, driver, x, y):
        for col_idx, col_size in enumerate(self.column_sizes[self._model.row_header_count():]):
            for header_idx in range(self._model.column_header_count()):
                title = str(self._model.get_column_header(col_idx, header_idx))
                driver.puts(x, y + header_idx, title[:col_size])
            x += col_size + self.spacing

    def _draw_row_headers(self, driver, x, y):
        for row_idx in range(self._model.row_count()):
            hx = x
            for header_idx in range(self._model.row_header_count()):
                col_size = self.column_sizes[header_idx]
                title = str(self._model.get_row_header(row_idx, header_idx))
                driver.puts(hx, y + row_idx, title[:col_size])
                hx += self.spacing + col_size

    def _draw_data(self, driver, x, y):
        for col_idx, col_size in enumerate(self.column_sizes[self._model.row_header_count():]):
            for row_idx in range(self._model.row_count()):
                title = str(self._model.get_item(row_idx, col_idx))
                driver.puts(x, y + row_idx, title[:col_size])
            x += col_size + self.spacing

    def on_draw(self, ev):
        # compute basic sizes
        row_header_count = self._model.row_header_count()
        row_header_skip = sum(self.column_sizes[:row_header_count]) + self.spacing * row_header_count
        column_header_skip = self._model.column_header_count()
        # draw
        ev.driver.pushcolor('normal')
        ev.driver.fill_clip()
        ev.driver.pushcolor('strong')
        self._draw_column_headers(ev.driver, ev.x + row_header_skip, ev.y)
        self._draw_row_headers(ev.driver, ev.x, ev.y + column_header_skip)
        ev.driver.popcolor()
        self._draw_data(ev.driver, ev.x + row_header_skip, ev.y + column_header_skip)
        ev.driver.popcolor()

    def on_keypress(self, ev):
        key_map = {
            'up': self.move_up,
            'down': self.move_down,
            'left': self.move_left,
            'right': self.move_right,
            'pageup':  self.move_pageup,
            'pagedown':  self.move_pagedown}
        if ev.keyname in key_map:
            key_map[ev.keyname]()
            self.redraw()
            return True

    def move_up(self):
        if self.acell.y > 0:
            self.acell.y -= 1
            self.spot.y = self.acell.y + self._model.column_header_count()
            return True
        return False

    def move_down(self):
        if self.acell.y < self._model.row_count() - 1:
            self.acell.y += 1
            self.spot.y = self.acell.y + self._model.column_header_count()
            return True
        return False

    def move_pageup(self):
        view_height = self.view_height - self._model.column_header_count()
        if self.acell.y >= view_height - 1:
            self.emit('scrollreq', - (view_height - 1))
            self.acell.y -= view_height - 1
            self.spot.y = self.acell.y + self._model.column_header_count()
        else:
            self.acell.y = 0
            self.spot.y = self._model.column_header_count()

    def move_pagedown(self):
        view_height = self.view_height - self._model.column_header_count()
        if self.acell.y <= self._model.row_count() - (view_height - 1):
            self.emit('scrollreq', (self.view_height - 1))
            self.acell.y += view_height - 1
            self.spot.y = self.acell.y + self._model.column_header_count()
        else:
            self.acell.y = self._model.row_count() - 1
            self.spot.y = self.acell.y + self._model.column_header_count()

    def move_left(self):
        if self.acell.x > 0:
            self.acell.x -= 1
            return True
        return False

    def move_right(self):
        if self.acell.x < len([col for col in self.columns if not col.header]) - 1:
            self.acell.x += 1
            return True
        return False

